"""Worker threads and tasks used by the Periscope viewer."""

from .utils import * # Imports QtCore, QThread, QObject, pyqtSignal, QRunnable,
                     # streamer, np, asyncio, time, socket, queue, traceback, sys,
                     # DEFAULT_AMPLITUDE, NETANAL_UPDATE_INTERVAL, DENSITY_GRID,
                     # SCATTER_POINTS, spectrum_from_slow_tod, pg,
                     # gaussian_filter, convolve, SMOOTH_SIGMA, LOG_COMPRESS,
                     # DEFAULT_MIN_FREQ, DEFAULT_MAX_FREQ, DEFAULT_NSAMPLES,
                     # DEFAULT_NPOINTS, DEFAULT_MAX_CHANNELS, DEFAULT_MAX_SPAN,
                     # DEFAULT_CABLE_LENGTH, concurrent, fitting (from rfmux.algorithms.measurement)

# fitting is already imported via 'from .utils import *' if utils.py imports it from rfmux.algorithms.measurement
# However, to be explicit for this module's direct dependency:
from rfmux.algorithms.measurement import fitting as fitting_module_direct # Alias to avoid conflict if utils also exports 'fitting'
from rfmux.algorithms.measurement import fitting_nonlinear # Import nonlinear fitting module
from rfmux.core.transferfunctions import exp_bin_noise_data # Import exponential binning function

# Additional imports for async fitting with ThreadPoolExecutor
import os
import concurrent.futures
from typing import Dict, Any, Optional, Callable
import platform

class UDPReceiver(QtCore.QThread):
    """
    Receives multicast packets in a dedicated QThread using the C++ ReadoutPacketReceiver.
    The C++ receiver handles packet reordering.
    """
    def __init__(self, host: str, module: int) -> None:
        super().__init__()
        self.module_id = module  # 1-indexed module ID (for Periscope)
        self.module_idx = module - 1  # 0-indexed (for packet filtering)

        # Create socket and C++ receiver
        # reorder_window=256: Maintain good packet reordering capability
        # queue_max_size=50000: Handle high data rates at FIR stage 0 (~38kHz)
        # flush_threshold=32: Flush every 32 packets for smooth updates (~54ms at FIR stage 6)
        self.sock = streamer.get_multicast_socket(host)
        self.receiver = streamer.ReadoutPacketReceiver(self.sock,
                                                       reorder_window=256,
                                                       queue_max_size=50000,
                                                       flush_threshold=16)

        # Queue reference will be set when we first see packets for our module
        self.queue = None
        self.serial = None

        # Statistics
        self.packets_received = 0
        self.packets_dropped = 0

    def get_dropped_packets(self):
        """Get cumulative dropped packet count from C++ queue statistics."""
        if self.queue is not None:
            stats = self.queue.get_stats()
            return stats.sequence_gaps + stats.packets_dropped
        return self.packets_dropped

    def get_received_packets(self):
        """Get cumulative received packet count from C++ queue statistics."""
        if self.queue is not None:
            stats = self.queue.get_stats()
            return stats.packets_received
        return self.packets_received

    def run(self):
        """Main reception loop - calls C++ receiver and retrieves packets."""
        while not self.isInterruptionRequested():
            try:
                # Call C++ receiver to read and process packets
                self.receiver.receive_batch(batch_size=16, timeout_ms=50)

                # Find our module's queue
                if self.queue is None:
                    # Look for a queue matching our module
                    for serial, module, q in self.receiver.get_all_queues():
                        if module == self.module_idx:
                            self.queue = q
                            self.serial = serial
                            print(f"[UDP] Found queue for serial={serial}, module={self.module_id}")
                            break

            except Exception as e:
                if self.isInterruptionRequested():
                    break
                print(f"[UDP] Error in receive loop: {e}")
                continue

    def stop(self):
        print(f"[UDP] UDP receiving thread stopped. Total packets received: {self.get_received_packets()}")
        print(f"[UDP] UDP receiving thread stopped. Total packets dropped: {self.get_dropped_packets()}")
        self.requestInterruption()
        try:
            self.sock.close()
        except OSError:
            pass

class IQSignals(QObject):
    done = pyqtSignal(int, str, object)

class IQTask(QRunnable):
    def __init__(self, row, ch, I, Q, dot_px, mode, signals: IQSignals):
        super().__init__()
        self.row = row; self.ch = ch; self.I = I.copy(); self.Q = Q.copy()
        self.dot_px = dot_px; self.mode = mode; self.signals = signals

    def run(self):
        if len(self.I) < 2: self._handle_insufficient_data(); return
        payload = self._compute_density() if self.mode == "density" else self._compute_scatter()
        self.signals.done.emit(self.row, self.mode, payload)
        
    def _handle_insufficient_data(self):
        # DENSITY_GRID from .utils
        empty_payload = (np.zeros((DENSITY_GRID, DENSITY_GRID), np.uint8), (0,1,0,1)) if self.mode == "density" else ([],[],[])
        self.signals.done.emit(self.row, self.mode, empty_payload)
        
    def _compute_density(self):
        # DENSITY_GRID, gaussian_filter, SMOOTH_SIGMA, LOG_COMPRESS, convolve from .utils
        g = DENSITY_GRID; hist = np.zeros((g, g), np.uint32)
        Imin, Imax = self.I.min(), self.I.max(); Qmin, Qmax = self.Q.min(), self.Q.max()
        if Imin == Imax or Qmin == Qmax: return (hist.astype(np.uint8), (Imin, Imax, Qmin, Qmax))
        ix = ((self.I - Imin) * (g - 1) / (Imax - Imin)).astype(np.intp)
        qy = ((self.Q - Qmin) * (g - 1) / (Qmax - Qmin)).astype(np.intp)
        np.add.at(hist, (qy, ix), 1)
        if self.dot_px > 1: self._apply_dot_dilation(hist, ix, qy, g)
        if gaussian_filter is not None and SMOOTH_SIGMA > 0:
            hist = gaussian_filter(hist.astype(np.float32), SMOOTH_SIGMA, mode="nearest")
        if LOG_COMPRESS: hist = np.log1p(hist, out=hist.astype(np.float32))
        if hist.max() > 0: hist = (hist * (255.0 / hist.max())).astype(np.uint8)
        return (hist, (Imin, Imax, Qmin, Qmax))
        
    def _apply_dot_dilation(self, hist, ix, qy, g):
        # convolve from .utils
        r = self.dot_px // 2
        if convolve is not None:
            k = 2 * r + 1; kernel = np.ones((k, k), dtype=np.uint8)
            hist[:] = convolve(hist, kernel, mode="constant", cval=0) # Update hist in place
        else:
            for dy in range(-r, r + 1):
                for dx in range(-r, r + 1):
                    ys, xs = qy + dy, ix + dx
                    mask = ((0 <= ys) & (ys < g) & (0 <= xs) & (xs < g))
                    np.add.at(hist, (ys[mask], xs[mask]), 1)
                    
    def _compute_scatter(self):
        # SCATTER_POINTS, pg from .utils
        N = len(self.I)
        idx = np.linspace(0, N - 1, SCATTER_POINTS, dtype=np.intp) if N > SCATTER_POINTS else np.arange(N, dtype=np.intp)
        xs, ys = self.I[idx], self.Q[idx]
        rel = idx / (idx.max() if idx.size else 1)
        colors = pg.colormap.get("turbo").map(rel.astype(np.float32), mode="byte")
        return (xs, ys, colors)

class PSDSignals(QObject):
    done = pyqtSignal(int, str, int, object)

class PSDTask(QRunnable):
    def __init__(self, row: int, ch: int, I: np.ndarray, Q: np.ndarray, mode: str, dec_stage: int,
                 real_units: bool, psd_absolute: bool, segments: int, signals: PSDSignals, exp_binning: bool = False, nbins: int = 1000):
        super().__init__()
        self.row, self.ch, self.I, self.Q, self.mode = row, ch, I.copy(), Q.copy(), mode
        self.dec_stage, self.real_units, self.psd_absolute = dec_stage, real_units, psd_absolute
        self.segments, self.signals = segments, signals
        self.exp_binning = exp_binning
        self.nbins = nbins

    def run(self):
        data_len = len(self.I)
        if data_len < 2: self._handle_insufficient_data(); return
        ref = "counts" if not self.real_units else ("absolute" if self.psd_absolute else "relative")
        nper = max(1, data_len // max(1, self.segments))
        # spectrum_from_slow_tod from .utils
        payload = self._compute_ssb_psd(ref, nper) if self.mode == "SSB" else self._compute_dsb_psd(ref, nper)
        self.signals.done.emit(self.row, self.mode, self.ch, payload)

    def _handle_insufficient_data(self):
        payload = ([], [], [], [], [], [], 0.0) if self.mode == "SSB" else ([], [])
        self.signals.done.emit(self.row, self.mode, self.ch, payload)
        
    def _compute_ssb_psd(self, ref, nper):
        # spectrum_from_slow_tod from .utils
        # Determine input units based on whether data was already converted to volts
        input_units = "volts" if self.real_units else "adc_counts"
        
        spec_iq = spectrum_from_slow_tod(i_data=self.I, q_data=self.Q, dec_stage=self.dec_stage,
                                         scaling="psd", reference=ref, nperseg=nper, spectrum_cutoff=0.9,
                                         input_units=input_units)
        
        freq_iq = spec_iq["freq_iq"]
        psd_i = spec_iq["psd_i"]
        psd_q = spec_iq["psd_q"]
        
        # For magnitude PSD: compute in frequency domain from I and Q PSDs
        # For uncorrelated I and Q noise, magnitude PSD ≈ PSD_I + PSD_Q
        # This avoids artifacts from computing PSD of time-domain magnitude
        if ref == "counts":
            # When in counts mode, PSDs are in linear scale
            psd_m = psd_i + psd_q
        else:
            # When in dB scale (dBc or dBm), convert to linear, add, then convert back
            psd_i_linear = 10**(psd_i / 10)
            psd_q_linear = 10**(psd_q / 10)
            psd_m_linear = psd_i_linear + psd_q_linear
            psd_m = 10 * np.log10(psd_m_linear)
        
        freq_m = freq_iq  # Same frequency grid
        
        # Apply exponential binning if enabled
        if self.exp_binning and len(freq_iq) > 1:
            freq_iq_binned, psd_i_binned = exp_bin_noise_data(freq_iq, psd_i, self.nbins)
            _, psd_q_binned = exp_bin_noise_data(freq_iq, psd_q, self.nbins)
            freq_m_binned, psd_m_binned = exp_bin_noise_data(freq_m, psd_m, self.nbins)
            return (freq_iq_binned, psd_i_binned, psd_q_binned, psd_m_binned,
                    freq_m_binned, psd_m_binned, float(self.dec_stage))
        
        return (freq_iq, psd_i, psd_q, psd_m, freq_m, psd_m, float(self.dec_stage))
        
    def _compute_dsb_psd(self, ref, nper):
        # spectrum_from_slow_tod from .utils
        # Determine input units based on whether data was already converted to volts
        input_units = "volts" if self.real_units else "adc_counts"
        
        spec_iq = spectrum_from_slow_tod(i_data=self.I, q_data=self.Q, dec_stage=self.dec_stage,
                                         scaling="psd", reference=ref, nperseg=nper, spectrum_cutoff=0.9,
                                         input_units=input_units)
        freq_dsb, psd_dsb = spec_iq["freq_dsb"], spec_iq["psd_dual_sideband"]
        order = np.argsort(freq_dsb)
        freq_dsb_sorted = freq_dsb[order]
        psd_dsb_sorted = psd_dsb[order]
        
        # Apply exponential binning if enabled
        if self.exp_binning and len(freq_dsb_sorted) > 1:
            freq_dsb_binned, psd_dsb_binned = exp_bin_noise_data(freq_dsb_sorted, psd_dsb_sorted, self.nbins)
            return (freq_dsb_binned, psd_dsb_binned)
        
        return (freq_dsb_sorted, psd_dsb_sorted)

class CRSInitializeSignals(QObject):
    success = pyqtSignal(str); error = pyqtSignal(str)

class NetworkAnalysisSignals(QObject):
    progress = pyqtSignal(int, float)
    data_update = pyqtSignal(int, np.ndarray, np.ndarray, np.ndarray)
    data_update_with_amp = pyqtSignal(int, np.ndarray, np.ndarray, np.ndarray, float)
    completed = pyqtSignal(int); error = pyqtSignal(str)

class DACScaleFetcher(QtCore.QThread):
    dac_scales_ready = QtCore.pyqtSignal(dict)
    def __init__(self, crs): super().__init__(); self.crs = crs
    def run(self):
        dac_scales = {}; loop = asyncio.new_event_loop(); asyncio.set_event_loop(loop)
        try: self._fetch_all_dac_scales(loop, dac_scales)
        finally: loop.close()
        self.dac_scales_ready.emit(dac_scales)
    def _fetch_all_dac_scales(self, loop, dac_scales):
        for module_idx in range(1, 9): # Renamed module
            try:
                dac_scale = loop.run_until_complete(self.crs.get_dac_scale('DBM', module=module_idx))
                if dac_scale is not None:
                    dac_scales[module_idx] = dac_scale - 1.5
                else:
                    dac_scales[module_idx] = None
            except Exception as e:
                if "Can't access module" in str(e) and "analog banking" in str(e):
                    dac_scales[module_idx] = None
                else:
                    print(f"Error fetching DAC scale for module {module_idx}: {e}", file=sys.stderr) # Print to stderr
                    dac_scales[module_idx] = None

class NetworkAnalysisTask(QtCore.QThread):
    """QThread subclass for performing network analysis operations without blocking the GUI."""
    def __init__(self, crs: "CRS", module: int, params: dict, signals: NetworkAnalysisSignals, amplitude=None):
        super().__init__()
        self.crs, self.module, self.params, self.signals = crs, module, params, signals
        # DEFAULT_AMPLITUDE, NETANAL_UPDATE_INTERVAL from .utils
        self.amplitude = amplitude if amplitude is not None else params.get('amp', DEFAULT_AMPLITUDE)
        self._running, self._last_update_time = True, 0
        self._update_interval = NETANAL_UPDATE_INTERVAL
        self._task, self._loop = None, None
        
    def stop(self):
        """Stop the network analysis task and cancel any ongoing async operation."""
        self._running = False
        self.requestInterruption()
    
    async def _cleanup_channels(self):
        try:
            # Direct approach to set amplitudes to zero without using async with
            for j in range(1, 1024):
                await self.crs.set_amplitude(0, channel=j, module=self.module)
        except Exception as e:
            print(f"Error in _cleanup_channels: {e}", file=sys.stderr)
            pass
    
    def run(self):
        """QThread entry point - runs in a separate thread."""
        # Create asyncio loop for this thread
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        
        try:
            progress_cb, data_cb = self._create_progress_callback(), self._create_data_callback()
            task_params = self._extract_parameters()
            
            # Setup phase: Clear channels and set cable length
            if task_params['clear_channels'] and not self.isInterruptionRequested():
                loop.run_until_complete(self.crs.clear_channels(module=self.module))
                
            if not self.isInterruptionRequested():
                loop.run_until_complete(self.crs.set_cable_length(length=task_params['cable_length'], module=self.module))
            
            # Main analysis phase
            if not self.isInterruptionRequested():
                # Combine parameters for the take_netanal call
                netanal_params = {
                    'amp': self.amplitude,
                    'fmin': task_params['fmin'],
                    'fmax': task_params['fmax'],
                    'nsamps': task_params['nsamps'],
                    'npoints': task_params['npoints'],
                    'max_chans': task_params['max_chans'],
                    'max_span': task_params['max_span'],
                    'module': self.module,
                    'progress_callback': progress_cb,
                    'data_callback': data_cb
                }
                
                # Process the network analysis asynchronously without blocking
                result = loop.run_until_complete(self._process_network_analysis(loop, netanal_params))
                
                # Process results if available and task wasn't interrupted
                if not self.isInterruptionRequested() and result:
                    fs_sorted, iq_sorted = result['frequencies'], result['iq_complex']
                    phase_sorted, amp_sorted = result['phase_degrees'], np.abs(iq_sorted)
                    self.signals.data_update.emit(self.module, fs_sorted, amp_sorted, phase_sorted)
                    self.signals.data_update_with_amp.emit(self.module, fs_sorted, amp_sorted, phase_sorted, self.amplitude)
                    self.signals.completed.emit(self.module)
            
        except asyncio.CancelledError:
            self.signals.error.emit(f"Analysis canceled for module {self.module}")
            if loop.is_running():
                loop.run_until_complete(self._cleanup_channels())
        except KeyError as ke:
            err_msg = f"KeyError accessing results for module {self.module}: {ke}. Expected 'frequencies', 'iq_complex', 'phase_degrees'."
            print(f"ERROR: {err_msg}", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            self.signals.error.emit(err_msg)
        except Exception as e:
            err_msg = f"Error processing results for module {self.module}: {type(e).__name__}: {e}"
            print(f"ERROR: {err_msg}", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            self.signals.error.emit(err_msg)
        finally:
            if loop.is_running():
                loop.stop()
            loop.close()
            
    def _create_progress_callback(self):
        return lambda module_idx, prog: self.signals.progress.emit(module_idx, prog) if self._running else None # Renamed module, progress
        
    def _create_data_callback(self):
        def data_cb(module_idx, freqs_raw, amps_raw, phases_raw): # Renamed module
            if self._running:
                sort_idx = np.argsort(freqs_raw)
                freqs, amps, phases = freqs_raw[sort_idx], amps_raw[sort_idx], phases_raw[sort_idx]
                current_time = time.time()
                if current_time - self._last_update_time >= self._update_interval:
                    self._last_update_time = current_time                    
                self.signals.data_update.emit(module_idx, freqs, amps, phases)
                self.signals.data_update_with_amp.emit(module_idx, freqs, amps, phases, self.amplitude)
        return data_cb
    
    def _extract_parameters(self):
        # Constants from .utils
        return {'fmin': self.params.get('fmin', DEFAULT_MIN_FREQ), 'fmax': self.params.get('fmax', DEFAULT_MAX_FREQ),
                'nsamps': self.params.get('nsamps', DEFAULT_NSAMPLES), 'npoints': self.params.get('npoints', DEFAULT_NPOINTS),
                'max_chans': self.params.get('max_chans', DEFAULT_MAX_CHANNELS), 'max_span': self.params.get('max_span', DEFAULT_MAX_SPAN),
                'cable_length': self.params.get('cable_length', DEFAULT_CABLE_LENGTH), 'clear_channels': self.params.get('clear_channels', True)}
        
    async def _process_network_analysis(self, loop, netanal_params):
        """Process a single network analysis operation asynchronously.
        
        This method periodically yields control back to the event loop to keep the GUI responsive.
        """
        netanal_coro = self.crs.take_netanal(**netanal_params)
        task = loop.create_task(netanal_coro)
        
        # Check for interruption while the task is running
        while not task.done():
            if self.isInterruptionRequested():
                task.cancel()
                await asyncio.sleep(0.01)  # Give the cancellation a chance to process
                return None
            await asyncio.sleep(0.1)  # Short sleep to yield control back to the event loop - this is crucial for preventing GUI freezing
        
        # Get the result when the task is done
        if not task.cancelled():
            try:
                return await task
            except Exception as e:
                print(f"Error in _process_network_analysis: {e}", file=sys.stderr)
                raise
        return None

class CRSInitializeTask(QRunnable):
    def __init__(self, crs: "CRS", module: int, irig_source: Any, clear_channels: bool, signals: CRSInitializeSignals):
        super().__init__(); self.crs, self.module, self.irig_source = crs, module, irig_source
        self.clear_channels, self.signals, self._loop = clear_channels, signals, None
    def run(self):
        # traceback from .utils
        self._loop = asyncio.new_event_loop(); asyncio.set_event_loop(self._loop)
        try:
            self._loop.run_until_complete(self._initialize_c_r_s())
            self.signals.success.emit("CRS board initialized successfully.")
        except Exception as e:
            self.signals.error.emit(f"Error during CRS initialization: {type(e).__name__}: {str(e)}\n{traceback.format_exc()}")
        finally:
            if self._loop:
                if self._loop.is_running(): self._loop.stop()
                self._loop.close(); self._loop = None
    async def _initialize_c_r_s(self):
        await self.crs.set_timestamp_port(self.irig_source)
        if self.clear_channels: await self.crs.clear_channels(module=self.module)

class SetCableLengthSignals(QObject):
    """Signals for SetCableLengthTask."""
    success = pyqtSignal(int, float)  # module_id, length_set
    error = pyqtSignal(int, str)    # module_id, error_message

class SetCableLengthTask(QRunnable):
    """A QRunnable task to set the cable length on the CRS asynchronously."""
    def __init__(self, crs: "CRS", module_id: int, length: float, signals: SetCableLengthSignals):
        super().__init__()
        self.crs = crs
        self.module_id = module_id
        self.length = length
        self.signals = signals
        self._loop = None

    def run(self):
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        try:
            self._loop.run_until_complete(self.crs.set_cable_length(length=self.length, module=self.module_id))
            self.signals.success.emit(self.module_id, self.length)
        except Exception as e:
            # traceback, sys are imported from .utils
            err_msg = f"Error setting cable length for module {self.module_id} to {self.length}m: {type(e).__name__}: {str(e)}"
            print(f"ERROR: {err_msg}", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            self.signals.error.emit(self.module_id, err_msg)
        finally:
            if self._loop:
                if self._loop.is_running():
                    self._loop.stop()
                self._loop.close()
            self._loop = None

class MultisweepSignals(QObject):
    progress = pyqtSignal(int, float)
    # Updated data_update to include iteration and direction:
    # 1. results_for_plotting: {output_cf: data_dict_val} - original structure from crs.multisweep
    # 2. results_for_history: {conceptual_idx: output_cf_key} - for easy history update
    data_update = pyqtSignal(int, int, float, str, dict, dict) # module, iteration, amplitude, direction, results_for_plotting, results_for_history
    completed_iteration = pyqtSignal(int, int, float, str) # module, iteration, amplitude, direction
    starting_iteration = pyqtSignal(int, int, float, str) # module, iteration, amplitude, direction
    fitting_progress = pyqtSignal(int, str) # module, status_message
    all_completed = pyqtSignal()
    error = pyqtSignal(int, float, str)

class MultisweepTask(QtCore.QThread):
    """QThread subclass for performing multisweep operations without blocking the GUI."""
    def __init__(self, crs: "CRS", params: dict, signals: MultisweepSignals, window: Any):
        """
        Initialize the MultisweepTask.
        
        Args:
            crs: Control and Readout System object
            params: Dictionary of parameters for the multisweep
            signals: Signal object for communication with GUI
            window: MultisweepWindow instance that will display the results
        """
        super().__init__()
        self.crs = crs
        self.params = params 
        self.signals = signals
        self.window = window 
        self.baseline_resonance_frequencies = list(params.get('resonance_frequencies', []))
        self._running = True
        self.current_amplitude = -1
        self.current_iteration = -1
        self.current_direction = ""
        self._task_completed = asyncio.Event()
        
        # ThreadPool configuration for fitting operations
        self._executor = None
        # Use multiple workers for parallel fitting (reserve 1 core for GUI)
        self._max_workers = max(1, min(4, (os.cpu_count() or 1) - 1))
        self._fitting_future = None

    def stop(self):
        self._running = False
        self.requestInterruption()

    def _progress_callback_wrapper(self, module_idx, progress_percentage):
        if self._running: self.signals.progress.emit(module_idx, progress_percentage)


    def run(self):
        """QThread entry point - runs in a separate thread."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        
        module_idx = self.params.get('module')
        if module_idx is None:
            self.signals.error.emit(-1, -1, "Module not specified in multisweep parameters.")
            return
        
        try:
            amplitudes = self.params.get('amps', [DEFAULT_AMPLITUDE])
            sweep_direction = self.params.get('sweep_direction', 'upward')
            conceptual_frequencies_from_window = self.window.conceptual_resonance_frequencies
            iteration_index = 0

            for amp_val in amplitudes:
                if self.isInterruptionRequested():
                    self.signals.error.emit(module_idx, self.current_amplitude, "Multisweep canceled.")
                    return
                
                self.current_amplitude = amp_val
                current_sweep_cfs_for_this_amp = []
                # conceptual_idx_to_input_cf_map = {} # Not strictly needed if results are always index-keyed

                if not conceptual_frequencies_from_window:
                    self.signals.error.emit(module_idx, amp_val, "Conceptual frequencies not available from window.")
                    return

                for idx, conceptual_cf in enumerate(conceptual_frequencies_from_window):
                    remembered_cf = self.window._get_closest_remembered_cf(idx, amp_val)
                    chosen_input_cf = remembered_cf if remembered_cf is not None else self.baseline_resonance_frequencies[idx]
                    current_sweep_cfs_for_this_amp.append(chosen_input_cf)
                    # conceptual_idx_to_input_cf_map[idx] = chosen_input_cf
                
                directions_to_sweep = ["upward","downward"] if sweep_direction == "both" else [sweep_direction]
                
                for direction_val in directions_to_sweep: # Renamed 'direction' to 'direction_val' to avoid conflict
                    self.current_iteration = iteration_index
                    self.current_direction = direction_val
                    
                    self.signals.starting_iteration.emit(module_idx, iteration_index, amp_val, direction_val)
                    
                    multisweep_params = {
                        'center_frequencies': current_sweep_cfs_for_this_amp,
                        'span_hz': self.params['span_hz'],
                        'npoints_per_sweep': self.params['npoints_per_sweep'],
                        'amp': amp_val,
                        'nsamps': self.params.get('nsamps', 10),
                        'module': module_idx,
                        'progress_callback': self._progress_callback_wrapper,
                        'bias_frequency_method': self.params.get('bias_frequency_method', 'max-diq'),
                        'rotate_saved_data': self.params.get('rotate_saved_data', False),
                        'sweep_direction': direction_val
                    }
                    
                    raw_results_from_crs = loop.run_until_complete(self._process_multisweep(loop, multisweep_params))
                    
                    if self.isInterruptionRequested():
                        self.signals.error.emit(module_idx, amp_val, "Multisweep canceled during execution.")
                        return
                    
                    # Process bifurcation detection first
                    if raw_results_from_crs:
                        for res_key, data_dict_val in raw_results_from_crs.items():
                            if not isinstance(res_key, (int, np.integer)): continue
                            iq_data = data_dict_val.get('iq_complex')
                            if iq_data is not None:
                                try:
                                    data_dict_val['is_bifurcated'] = fitting_module_direct.identify_bifurcation(iq_data, threshold_factor=7)
                                except Exception as e:
                                    print(f"Warning: Bifurcation detection failed for index {res_key}: {e}", file=sys.stderr)
                                    data_dict_val['is_bifurcated'] = False
                            else:
                                data_dict_val['is_bifurcated'] = False
                    
                    # Now apply fitting analysis using the (potentially) bifurcation-annotated data
                    # Use async version with ThreadPoolExecutor for better responsiveness
                    enhanced_results = loop.run_until_complete(
                        self._process_fitting_async(raw_results_from_crs)
                    )
                    results_for_plotting = enhanced_results 
                    
                    results_for_history = {}
                    if enhanced_results: 
                        for res_idx, data_dict_val in enhanced_results.items():
                            if isinstance(res_idx, (int, np.integer)):
                                bias_freq = data_dict_val.get('bias_frequency', data_dict_val.get('original_center_frequency'))
                                if bias_freq is not None:
                                    # Convert 1-based detector index to 0-based conceptual index
                                    conceptual_idx = res_idx - 1
                                    results_for_history[conceptual_idx] = bias_freq
                                else: print(f"Warning: No bias frequency found for index {res_idx}", file=sys.stderr)
                            else: print(f"Warning (MultisweepTask): Non-integer key {res_idx} in results", file=sys.stderr)
                    
                    if results_for_plotting is not None:
                        self.signals.data_update.emit(module_idx, iteration_index, amp_val, direction_val, results_for_plotting, results_for_history)
                    
                    self.signals.completed_iteration.emit(module_idx, iteration_index, amp_val, direction_val)
                    iteration_index += 1
            
            self.signals.all_completed.emit()
        except asyncio.CancelledError:
            self.signals.error.emit(module_idx, self.current_amplitude, "Multisweep task canceled by user.")
        except Exception as e:
            detailed_error = f"Error in MultisweepTask (amp {self.current_amplitude:.4f}): {type(e).__name__}: {str(e)}\n{traceback.format_exc()}"
            self.signals.error.emit(module_idx, self.current_amplitude, detailed_error)
        finally:
            if loop.is_running():
                loop.stop()
            loop.close()

    async def _process_multisweep(self, loop, multisweep_params):
        self._task_completed.clear()
        multisweep_coro = self.crs.multisweep(**multisweep_params)
        task = loop.create_task(multisweep_coro)
        
        while not task.done():
            if self.isInterruptionRequested():
                task.cancel()
                await asyncio.sleep(0.01) 
                return None
            await asyncio.sleep(0.1) 
        
        if not task.cancelled():
            try:
                return await task
            except Exception as e:
                print(f"Error in _process_multisweep: {e}", file=sys.stderr)
                raise
        return None
    
    async def _process_fitting_async(self, raw_results):
        """Process fitting analysis asynchronously using ThreadPoolExecutor."""
        if not raw_results:
            return raw_results
        
        # For larger datasets, use parallel processing
        return await self._process_fitting_multi_thread(raw_results)
    
    async def _process_fitting_multi_thread(self, raw_results):
        """Process fitting using multiple threads for better performance."""
        loop = asyncio.get_event_loop()
        
        # Split resonances into chunks for parallel processing
        resonance_items = list(raw_results.items())
        num_chunks = min(self._max_workers, len(resonance_items))
        chunk_size = max(1, len(resonance_items) // num_chunks)
        
        chunks = []
        for i in range(0, len(resonance_items), chunk_size):
            chunk = dict(resonance_items[i:i + chunk_size])
            if chunk:
                chunks.append(chunk)
        
        # Process chunks in parallel
        with concurrent.futures.ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            # Submit all chunks for processing
            futures = []
            for i, chunk in enumerate(chunks):
                future = executor.submit(
                    self._process_chunk_thread_safe,
                    chunk,
                    self.params.get('apply_skewed_fit', False),
                    self.params.get('apply_nonlinear_fit', False),
                    self.params.get('module'),
                    i,
                    len(chunks)
                )
                futures.append(asyncio.wrap_future(future))
            
            # Collect results as they complete
            enhanced_results = {}
            completed = 0
            
            for future in asyncio.as_completed(futures):
                if self.isInterruptionRequested():
                    # Cancel remaining futures
                    for f in futures:
                        f.cancel()
                    return raw_results
                
                try:
                    chunk_results = await future
                    enhanced_results.update(chunk_results)
                    completed += 1
                    
                    # Chunk completed (no progress emission to avoid clutter)
                        
                except Exception as e:
                    print(f"Error processing chunk: {e}", file=sys.stderr)
                    # Continue with other chunks
            
            return enhanced_results
    
    def _process_chunk_thread_safe(self, chunk, apply_skewed, apply_nonlinear, module_idx, chunk_idx, total_chunks):
        """Process a chunk of resonances in a thread."""
        # Process this chunk using the existing logic
        return self._apply_fitting_analysis(chunk, apply_skewed, apply_nonlinear, module_idx)
    
    def _apply_fitting_analysis(self, raw_results, apply_skewed, apply_nonlinear, module_idx):
        """Apply fitting analysis (skewed and/or nonlinear) to multisweep results.
        
        This method runs in a worker thread for parallel processing.
        Emits progress signals for GUI updates (thread-safe via Qt's queued connections).
        """
        # This runs in a separate thread, so we need to be careful about Qt signals
        
        # Initialize enhanced_results as a copy to preserve raw data if fitting is skipped or fails
        enhanced_results = {k: v.copy() for k, v in raw_results.items()}
        
        if not apply_skewed and not apply_nonlinear:
            # If no fitting is requested, add flags indicating this
            for res_idx in enhanced_results:
                enhanced_results[res_idx]['skewed_fit_applied'] = False
                enhanced_results[res_idx]['skewed_fit_success'] = False
                enhanced_results[res_idx]['nonlinear_fit_applied'] = False
                enhanced_results[res_idx]['nonlinear_fit_success'] = False
            return enhanced_results
        
        try:
            if apply_skewed:
                # Emit progress signal (thread-safe via Qt's queued connections)
                if module_idx is not None and self._running:
                    self.signals.fitting_progress.emit(module_idx, 
                        "Fitting in progress: Applying skewed fits")
                
                # Perform skewed fitting
                skewed_results = fitting_module_direct.fit_skewed_multisweep(
                    enhanced_results,
                    approx_Q_for_fit=1e4,
                    fit_resonances=True,
                    center_iq_circle=True,
                    normalize_fit=True
                )
                
                # Update results
                for res_idx in enhanced_results:
                    if res_idx in skewed_results:
                        enhanced_results[res_idx].update(skewed_results[res_idx])
                        enhanced_results[res_idx]['skewed_fit_applied'] = True
                        fit_p = enhanced_results[res_idx].get('fit_params', {})
                        enhanced_results[res_idx]['skewed_fit_success'] = fit_p.get('fr') is not None and fit_p.get('fr') != 'nan'
                        
                        # Generate skewed model magnitude if fit was successful
                        if enhanced_results[res_idx]['skewed_fit_success'] and fit_p:
                            frequencies = enhanced_results[res_idx].get('frequencies')
                            if frequencies is not None:
                                try:
                                    # Generate magnitude model using fitted parameters
                                    skewed_model_mag = fitting_module_direct.s21_skewed(
                                        frequencies, 
                                        fit_p['fr'], 
                                        fit_p['Qr'], 
                                        fit_p['Qcre'], 
                                        fit_p['Qcim'], 
                                        fit_p['A']
                                    )
                                    enhanced_results[res_idx]['skewed_model_mag'] = skewed_model_mag
                                except Exception as e:
                                    print(f"Warning: Failed to generate skewed model for resonance {res_idx}: {e}", file=sys.stderr)
                    else:
                        enhanced_results[res_idx]['skewed_fit_applied'] = True
                        enhanced_results[res_idx]['skewed_fit_success'] = False
            else:
                for res_idx in enhanced_results:
                    enhanced_results[res_idx]['skewed_fit_applied'] = False
                    enhanced_results[res_idx]['skewed_fit_success'] = False
            
            if apply_nonlinear:
                # Emit progress signal
                if module_idx is not None and self._running:
                    self.signals.fitting_progress.emit(module_idx, 
                        "Fitting in progress: Applying non-linear fits")
                
                # Perform nonlinear fitting
                # Disable parallel processing since we're already in a thread
                nonlinear_results = fitting_nonlinear.fit_nonlinear_iq_multisweep(
                    enhanced_results.copy(),
                    fit_nonlinearity=True,
                    n_extrema_points=5,
                    verbose=False,
                    parallel=False  # Avoid nested thread pools
                )
                
                # Update results
                for res_idx in enhanced_results:
                    if res_idx in nonlinear_results:
                        enhanced_results[res_idx].update(nonlinear_results[res_idx])
                        enhanced_results[res_idx]['nonlinear_fit_applied'] = True
                        if 'nonlinear_fit_success' not in enhanced_results[res_idx]:
                            enhanced_results[res_idx]['nonlinear_fit_success'] = False
                        
                        # Generate nonlinear model IQ if fit was successful
                        if enhanced_results[res_idx].get('nonlinear_fit_success', False):
                            nl_params = enhanced_results[res_idx].get('nonlinear_fit_params', {})
                            frequencies = enhanced_results[res_idx].get('frequencies')
                            if nl_params and frequencies is not None:
                                try:
                                    # Generate complex IQ model using fitted parameters
                                    nonlinear_model_iq = fitting_nonlinear.nonlinear_iq(
                                        frequencies,
                                        nl_params['fr'],
                                        nl_params['Qr'],
                                        nl_params['amp'],
                                        nl_params['phi'],
                                        nl_params['a'],
                                        nl_params['i0'],
                                        nl_params['q0']
                                    )
                                    enhanced_results[res_idx]['nonlinear_model_iq'] = nonlinear_model_iq
                                except Exception as e:
                                    print(f"Warning: Failed to generate nonlinear model for resonance {res_idx}: {e}", file=sys.stderr)
                    else:
                        enhanced_results[res_idx]['nonlinear_fit_applied'] = True
                        enhanced_results[res_idx]['nonlinear_fit_success'] = False
            else:
                for res_idx in enhanced_results:
                    enhanced_results[res_idx]['nonlinear_fit_applied'] = False
                    enhanced_results[res_idx]['nonlinear_fit_success'] = False
            
            # Emit completion
            if module_idx is not None and self._running:
                self.signals.fitting_progress.emit(module_idx, "Fitting Completed")
            
            return enhanced_results
            
        except Exception as e:
            print(f"Error in thread-safe fitting analysis: {e}", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            # Return results with error flags
            for res_idx in enhanced_results:
                enhanced_results[res_idx]['fitting_error'] = str(e)
            return enhanced_results


class BiasKidsSignals(QObject):
    """Signals for BiasKidsTask."""
    progress = pyqtSignal(int, float)  # module, progress_percentage
    completed = pyqtSignal(int, dict, dict, float)  # module, biased_results, df_calibrations, nco_frequency_hz
    error = pyqtSignal(str)  # error_message


class BiasKidsTask(QtCore.QThread):
    """QThread subclass for running the bias_kids algorithm without blocking the GUI."""
    
    def __init__(self, crs: "CRS", module: int, multisweep_results: dict, signals: BiasKidsSignals, bias_params: Optional[Dict[str, Any]] = None):
        """
        Initialize the BiasKidsTask.
        
        Args:
            crs: Control and Readout System object
            module: Module number to bias
            multisweep_results: Multisweep results in GUI format
            signals: Signal object for communication with GUI
            bias_params: Optional dictionary of bias parameters from dialog
        """
        super().__init__()
        self.crs = crs
        self.module = module
        self.multisweep_results = multisweep_results
        self.signals = signals
        self.bias_params = bias_params or {}
        self._running = True
        
    def stop(self):
        """Stop the task."""
        self._running = False
        self.requestInterruption()
        
    def run(self):
        """QThread entry point - runs in a separate thread."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        
        try:
            # Progress callback
            def progress_cb(module, progress):
                if self._running:
                    self.signals.progress.emit(module, progress)
            
            # Run the bias_kids algorithm
            result = loop.run_until_complete(self._run_bias_kids(progress_cb))
            
            if self.isInterruptionRequested():
                self.signals.error.emit("Bias KIDs operation was cancelled.")
                return
                
            if result:
                # Handle both dict and list return types from bias_kids
                if isinstance(result, list):
                    # For list results (multiple modules), we're only processing one module here
                    # so this shouldn't happen, but handle it gracefully
                    if len(result) > 0:
                        result = result[0]  # Take the first module's results
                    else:
                        self.signals.error.emit("Bias KIDs operation returned empty list.")
                        return
                
                # Extract df_calibration values (result is now guaranteed to be a dict)
                df_calibrations = {}
                for det_idx, det_data in result.items():
                    if 'df_calibration' in det_data:
                        df_calibrations[det_idx] = det_data['df_calibration']
                
                # Read the NCO frequency that was used during biasing
                nco_frequency_hz = loop.run_until_complete(self.crs.get_nco_frequency(module=self.module))
                
                # Emit completion with results and NCO frequency
                self.signals.completed.emit(self.module, result, df_calibrations, float(nco_frequency_hz))
            else:
                self.signals.error.emit("Bias KIDs operation returned no results.")
                
        except asyncio.CancelledError:
            self.signals.error.emit("Bias KIDs operation was cancelled.")
        except Exception as e:
            import traceback
            error_msg = f"Error in BiasKidsTask: {type(e).__name__}: {str(e)}\n{traceback.format_exc()}"
            print(error_msg)  # Print detailed message to Console
            self.signals.error.emit(str(e))
        finally:
            if loop.is_running():
                loop.stop()
            loop.close()
            
    async def _run_bias_kids(self, progress_callback):
        """Run the bias_kids algorithm asynchronously."""
        # Import bias_kids as a regular function
        from rfmux.algorithms.measurement.bias_kids import bias_kids
        
        # Extract parameters from bias_params
        kwargs = {
            'crs': self.crs,
            'multisweep_results': self.multisweep_results,
            'module': self.module,
            'progress_callback': progress_callback
        }
        
        # Add optional parameters from dialog
        if 'nonlinear_threshold' in self.bias_params:
            kwargs['nonlinear_threshold'] = self.bias_params['nonlinear_threshold']
        if 'fallback_to_lowest' in self.bias_params:
            kwargs['fallback_to_lowest'] = self.bias_params['fallback_to_lowest']
        if 'optimize_phase' in self.bias_params:
            kwargs['optimize_phase'] = self.bias_params['optimize_phase']
        if 'bandpass_params' in self.bias_params:
            kwargs['bandpass_params'] = self.bias_params['bandpass_params']
        if 'num_phase_samples' in self.bias_params:
            kwargs['num_phase_samples'] = self.bias_params['num_phase_samples']
        if 'phase_step' in self.bias_params:
            kwargs['phase_step'] = self.bias_params['phase_step']
        
        # Call bias_kids with all parameters
        result = await bias_kids(**kwargs)
        return result



"""
This module is derived from and based on the mr_resonator project by Maclean Rouble:
https://github.com/macleaner/mr_resonator

Modifications in this repository include:
- Trimmed to include only functions used by the Mock CRS framework
- Consolidation and integration with local JIT physics approximations

Original project license: see rfmux/mr_resonator/LICENSE (upstream LICENSE retained)

JIT-compiled physics calculations for LEKID resonator simulations.

This module consolidates all numba-accelerated physics calculations,
providing maximum performance for multi-resonator simulations.
All functions are JIT-compiled for 10-25x speedup over pure Python.

Numba is a required dependency for this module.
"""
import numpy as np
from numba import jit, prange
import os
import platform
import subprocess
import numba

# Physical constants
H = 6.626e-34  # Planck constant
KB = 1.38e-23  # Boltzmann constant
MU0 = 8.85e-12  # Permeability (note: this appears to be using permittivity value from original)

# ============================================================================
# Bessel Function Approximations
# ============================================================================

if platform.system() == "Darwin":
    numba.get_num_threads()
    layer = numba.config.THREADING_LAYER

    if layer != "omp":
        print(">>>>>>>>>> You are on Mac with no libomp, the processing will fail <<<<<<<<<<")
        print(">>>>>>>>>> Consult the README.MD <<<<<<<<<<<<<<\n")
    else:
        print("MacOS numba threading layer:", layer)


@jit(nopython=True, cache=True, fastmath=True)
def bessel_k0(x):
    """
    Fast approximation of modified Bessel function K0.
    
    Accurate to ~1e-6 relative error for all x > 0.
    """
    if x < 0.001:
        return 15.0 - np.log(x/2)
    elif x < 2.0:
        log_term = np.log(x / 2.0)
        gamma = 0.5772156649
        k0 = -log_term - gamma
        k0 += 0.25 * x * x * (log_term + gamma - 0.5)
        k0 += 0.015625 * x**4 * (log_term + gamma - 0.75)
        return k0
    else:
        sqrt_term = np.sqrt(np.pi / (2.0 * x))
        exp_term = np.exp(-x)
        series = 1.0 + 1.0/(8.0*x) + 9.0/(128.0*x*x)
        return sqrt_term * exp_term * series


@jit(nopython=True, cache=True, fastmath=True)
def bessel_i0(x):
    """
    Fast approximation of modified Bessel function I0.
    
    Accurate to ~1e-6 relative error for all x.
    """
    if x < 0.0:
        x = -x
    if x < 3.75:
        t = x / 3.75
        t2 = t * t
        return 1.0 + 3.5156229*t2 + 3.0899424*t2*t2 + 1.2067492*t2*t2*t2 + \
               0.2659732*t2*t2*t2*t2 + 0.0360768*t2*t2*t2*t2*t2 + 0.0045813*t2*t2*t2*t2*t2*t2
    else:
        t = 3.75 / x
        exp_term = np.exp(x)
        sqrt_term = 1.0 / np.sqrt(2.0 * np.pi * x)
        series = 0.39894228 + 0.01328592*t + 0.00225319*t*t - 0.00157565*t*t*t + \
                 0.00916281*t*t*t*t - 0.02057706*t*t*t*t*t
        return exp_term * series * sqrt_term


# ============================================================================
# Physics Calculations
# ============================================================================

@jit(nopython=True, cache=True)
def calc_sigma1(f, T, nqp, Delta0, N0, sigmaN):
    """
    Calculate real part of complex conductivity (sigma1).
    
    Parameters
    ----------
    f : float
        Frequency in Hz
    T : float  
        Temperature in K
    nqp : float
        Quasiparticle density
    Delta0 : float
        Zero-temperature gap energy
    N0 : float
        Density of states
    sigmaN : float
        Normal conductivity
        
    Returns
    -------
    float
        Real part of complex conductivity
    """
    zeta = H * f / (2.0 * KB * T)
    K0 = bessel_k0(zeta)
    
    x1 = 2.0 * Delta0 / (H * f)
    x2 = nqp / (N0 * np.sqrt(2.0 * np.pi * KB * T * Delta0))
    
    return x1 * x2 * np.sinh(zeta) * K0 * sigmaN


@jit(nopython=True, cache=True)
def calc_sigma2(f, T, nqp, Delta0, N0, sigmaN):
    """
    Calculate imaginary part of complex conductivity (sigma2).
    
    Parameters
    ----------
    f : float
        Frequency in Hz
    T : float
        Temperature in K
    nqp : float
        Quasiparticle density
    Delta0 : float
        Zero-temperature gap energy
    N0 : float
        Density of states
    sigmaN : float
        Normal conductivity
        
    Returns
    -------
    float
        Imaginary part of complex conductivity
    """
    zeta = H * f / (2.0 * KB * T)
    I0 = bessel_i0(zeta)
    
    x1 = np.pi * Delta0 / (H * f)
    x2 = nqp / (2.0 * N0 * Delta0)
    x3 = np.sqrt(2.0 * Delta0 / (np.pi * KB * T)) * np.exp(-zeta) * I0
    
    return x1 * (1.0 - x2 * (1.0 + x3)) * sigmaN


@jit(nopython=True, cache=True)
def calc_Zs(f, sigma1, sigma2, thickness, width, length):
    """
    Calculate surface impedance.
    
    Parameters
    ----------
    f : float
        Frequency in Hz
    sigma1, sigma2 : float
        Real and imaginary parts of conductivity
    thickness, width, length : float
        Geometric parameters in meters
        
    Returns
    -------
    complex
        Surface impedance
    """
    sigma = sigma1 - 1j * sigma2
    root1 = (1j * 2.0 * np.pi * f * MU0) / sigma
    cotharg = thickness * np.sqrt(1j * 2.0 * np.pi * f * MU0 * sigma)
    Zs = np.sqrt(root1) * (1.0 / np.tanh(cotharg))
    return Zs


@jit(nopython=True, cache=True)
def calc_R_L(f, Zs, length, width, R_spoiler):
    """
    Calculate total resistance and kinetic inductance.
    
    Parameters
    ----------
    f : float
        Frequency in Hz
    Zs : complex
        Surface impedance
    length, width : float
        Geometric parameters in meters
    R_spoiler : float
        Additional resistance
        
    Returns
    -------
    tuple of float
        (R_total, Lk_total)
    """
    R = Zs.real * (length / width) + R_spoiler
    L = (Zs.imag / (2.0 * np.pi * f)) * (length / width)
    return R, L


# ============================================================================
# Vectorized Physics Operations
# ============================================================================

@jit(nopython=True, parallel=True, cache=True)
def vectorized_update_params_from_nqp(
    nqp_array, readout_freqs, T_array, Delta0_array, 
    N0_array, sigmaN_array, thickness_array, 
    width_array, length_array, R_spoiler_array
):
    """
    Update R and Lk for all resonators based on quasiparticle density.
    
    Uses parallel execution for maximum performance with many resonators.
    
    Parameters
    ----------
    nqp_array : ndarray
        Quasiparticle densities for all resonators
    readout_freqs : ndarray
        Readout frequencies for all resonators
    T_array : ndarray
        Temperature for each resonator
    Delta0_array : ndarray
        Gap energy for each resonator
    N0_array : ndarray
        Density of states for each resonator
    sigmaN_array : ndarray
        Normal conductivity for each resonator
    thickness_array : ndarray
        Thickness for each resonator
    width_array : ndarray
        Width for each resonator
    length_array : ndarray
        Length for each resonator
    R_spoiler_array : ndarray
        Spoiler resistance for each resonator
    
    Returns
    -------
    R_array, Lk_array : ndarray
        Resistance and kinetic inductance for all resonators
    """
    n = len(nqp_array)
    R_out = np.zeros(n, dtype=np.float64)
    Lk_out = np.zeros(n, dtype=np.float64)
    
    # Parallel loop over all resonators
    for i in prange(n):
        # Calculate conductivities
        sigma1 = calc_sigma1(
            readout_freqs[i], T_array[i], nqp_array[i],
            Delta0_array[i], N0_array[i], sigmaN_array[i]
        )
        sigma2 = calc_sigma2(
            readout_freqs[i], T_array[i], nqp_array[i],
            Delta0_array[i], N0_array[i], sigmaN_array[i]
        )
        
        # Calculate surface impedance
        Zs = calc_Zs(
            readout_freqs[i], sigma1, sigma2,
            thickness_array[i], width_array[i], length_array[i]
        )
        
        # Calculate R and Lk
        R, Lk = calc_R_L(
            readout_freqs[i], Zs, length_array[i], 
            width_array[i], R_spoiler_array[i]
        )
        
        R_out[i] = R
        Lk_out[i] = Lk
    
    return R_out, Lk_out


# ============================================================================
# Convergence Loop
# ============================================================================

@jit(nopython=True, parallel=True, cache=True, fastmath=True)
def converged_lekid_parameters(
    frequency, amplitude, 
    L_array, R_array, C_array, Cc_array,
    base_Lk, base_Lg, base_L_junk,
    input_atten_dB, ZLNA,
    Istar, tolerance, max_iterations, damp=0.1
):
    """
    Self-consistent convergence loop for current-dependent inductance.
    
    Performs the entire convergence calculation in compiled code for
    maximum performance (2-5x speedup over Python loops).
    
    Parameters
    ----------
    frequency : float
        Probe frequency in Hz
    amplitude : float
        Input voltage amplitude
    L_array, R_array, C_array, Cc_array : ndarray
        Initial circuit parameters for all resonators.
        L_array = Lk + Lg + L_junk (total resonator inductance)
    base_Lk, base_Lg, base_L_junk : ndarray
        Base inductance values (before current modification).
        Only Lk changes with current; Lg and L_junk are fixed.
    input_atten_dB : float
        Input attenuation in dB
    ZLNA : complex
        LNA impedance
    Istar : float
        Characteristic current
    tolerance : float
        Convergence tolerance
    max_iterations : int
        Maximum convergence iterations
    damp : float
        Damping factor for convergence stability
        
    Returns
    -------
    L_converged : ndarray
        Converged total inductance values (Lk + Lg + L_junk)
    R_converged : ndarray
        Converged resistance values (unchanged)
    currents_converged : ndarray (complex)
        Converged resonator currents
    iterations : int
        Number of iterations to convergence
    """
    n = len(L_array)
    w = 2.0 * np.pi * frequency
    
    # Working arrays (copies to avoid modifying inputs)
    L_work = L_array.copy()
    currents_array = np.zeros(n, dtype=np.complex128)
    current_factors = np.ones(n, dtype=np.float64)
    
    # Attenuator values
    att_factor = 10.0**(input_atten_dB/20.0)
    z0 = 50.0  # Characteristic impedance
    r1 = z0 * ((att_factor + 1) / (att_factor - 1))
    r3 = r1
    r2 = (z0 / 2.0) * ((10.0**(input_atten_dB/10.0) - 1) / att_factor)
    
    actual_iterations = 0
    
    # Convergence loop
    for iteration in range(max_iterations):
        # Step 1: Calculate impedances for all resonators
        impedances = np.zeros(n, dtype=np.complex128)
        
        for i in prange(n):
            # Parallel RLC impedance
            # Note: L_work already includes L_junk (total resonator inductance)
            if C_array[i] > 0:
                ZC = 1.0 / (1j * w * C_array[i])
                ZL = 1j * w * L_work[i]
                Z_parallel_inv = 1.0/ZC + 1.0/(ZL + R_array[i])
                Z_parallel = 1.0 / Z_parallel_inv
            else:
                Z_parallel = 1j * w * L_work[i] + R_array[i]
            
            # Series coupling capacitor
            ZCc = 1.0 / (1j * w * Cc_array[i])
            
            # L_junk is now included in L_work, not added separately
            impedances[i] = Z_parallel + ZCc
        
        # Step 2: Calculate currents through resonators
        currents_new = np.zeros(n, dtype=np.complex128)
        
        for i in prange(n):
            # System impedance (parallel combination of resonator and LNA)
            Zsys = 1.0 / (1.0/impedances[i] + 1.0/ZLNA)
            
            # P-type attenuator calculation
            Zp = 1.0 / (1.0/Zsys + 1.0/r3)
            
            # Current through r2
            I2 = amplitude / (r2 + Zp)
            
            # Current divider for input current
            Iin = I2 * (r3 / (Zsys + r3))
            
            # Current through resonator (current divider)
            Zpar = 1.0 / (1.0/r3 + 1.0/impedances[i] + 1.0/ZLNA)
            currents_new[i] = Iin * Zpar / impedances[i]
        
        # Step 3: Apply damping for stability
        currents_array = currents_array + damp * (currents_new - currents_array)
        
        # Step 4: Calculate new current factors
        new_factors = 1.0 + (np.abs(currents_array)**2 / (Istar * Istar))
        
        # Step 5: Check convergence
        if iteration > 0:
            factor_change = np.max(np.abs(new_factors - current_factors))
            
            # Early stop after minimum iterations
            if iteration >= 3 and factor_change < tolerance:
                actual_iterations = iteration + 1
                break
            
            # Strict convergence
            if factor_change < tolerance * 0.1:
                actual_iterations = iteration + 1
                break
        
        # Step 6: Update factors and inductances
        current_factors = new_factors
        
        # Update L values: Lk changes with current; Lg and L_junk are fixed
        for i in prange(n):
            Lk_work = base_Lk[i] * current_factors[i]
            L_work[i] = Lk_work + base_Lg[i] + base_L_junk[i]
    
    if actual_iterations == 0:
        actual_iterations = max_iterations
    
    # Return converged values
    return L_work, R_array, currents_array, actual_iterations


# ============================================================================
# S21 Calculation
# ============================================================================

@jit(nopython=True, parallel=True, cache=True, fastmath=True)
def compute_s21_vectorized(
    fc, Vin,
    L_array, C_array, R_array, Cc_array,
    ZLNA, GLNA, input_atten_dB, system_termination
):
    """
    Vectorized S21 calculation for all resonators.
    
    Computes the voltage response (Vout) for multiple resonators in parallel,
    which can be converted to S21 by dividing by Vin.
    
    Parameters
    ----------
    fc : float
        Probe frequency in Hz
    Vin : float
        Input voltage amplitude
    L_array : ndarray
        Total inductance for each resonator (Lk + Lg + L_junk)
    C_array : ndarray
        Capacitance for each resonator
    R_array : ndarray
        Resistance for each resonator
    Cc_array : ndarray
        Coupling capacitance for each resonator
    ZLNA : complex
        LNA impedance
    GLNA : float
        LNA gain
    input_atten_dB : float
        Input attenuation in dB
    system_termination : float
        System termination impedance
        
    Returns
    -------
    ndarray (complex)
        Vout for each resonator
    """
    n = len(L_array)
    w = 2.0 * np.pi * fc
    Vout_array = np.zeros(n, dtype=np.complex128)
    
    # Calculate attenuator values
    att_factor = 10.0**(input_atten_dB/20.0)
    z0 = system_termination
    r1 = z0 * ((att_factor + 1) / (att_factor - 1))
    r3 = r1
    r2 = (z0 / 2.0) * ((10.0**(input_atten_dB/10.0) - 1) / att_factor)
    
    for i in prange(n):
        # Calculate impedance for this resonator
        # L_array already includes L_junk (total resonator inductance)
        if C_array[i] > 0:
            ZC = 1.0 / (1j * w * C_array[i])
            ZL = 1j * w * L_array[i]
            Z_parallel_inv = 1.0/ZC + 1.0/(ZL + R_array[i])
            Z_parallel = 1.0 / Z_parallel_inv
        else:
            Z_parallel = 1j * w * L_array[i] + R_array[i]
        
        # Add coupling capacitor
        ZCc = 1.0 / (1j * w * Cc_array[i])
        
        # Total resonator impedance (L_junk already in L_array)
        Z_res = Z_parallel + ZCc
        
        # System impedance calculation
        Zsys = 1.0 / (1.0/Z_res + 1.0/ZLNA)
        
        # P-type attenuator calculation
        Zp = 1.0 / (1.0/Zsys + 1.0/r3)
        
        # Current through r2
        I2 = Vin / (r2 + Zp)
        
        # Current divider for input current
        Iin = I2 * (r3 / (Zsys + r3))
        
        # Current through resonator
        Zpar = 1.0 / (1.0/r3 + 1.0/Z_res + 1.0/ZLNA)
        Ires = Iin * Zpar / Z_res
        
        # Voltage across parallel LC (at top of resonator)
        Vtop = Ires * Z_parallel
        
        # Apply LNA gain
        Vout_array[i] = Vtop * GLNA
    
    return Vout_array


# ============================================================================
# Parallel Resonator S21 Calculation
# ============================================================================

@jit(nopython=True, cache=True, fastmath=True)
def compute_s21_parallel(
    fc, Vin,
    L_array, C_array, R_array, Cc_array,
    ZLNA, GLNA, input_atten_dB, system_termination
):
    """
    Calculate S21 for multiple resonators in parallel on a transmission line.
    
    This properly combines all resonators in parallel before calculating
    the transmission to the load, giving physically correct dips at resonance.
    
    Parameters
    ----------
    fc : float
        Probe frequency in Hz
    Vin : float
        Input voltage amplitude
    L_array : ndarray
        Total inductance for each resonator (Lk + Lg + L_junk)
    C_array : ndarray
        Capacitance for each resonator
    R_array : ndarray
        Resistance for each resonator
    Cc_array : ndarray
        Coupling capacitance for each resonator
    ZLNA : complex
        LNA impedance (load at end of transmission line)
    GLNA : float
        LNA gain
    input_atten_dB : float
        Input attenuation in dB
    system_termination : float
        System termination impedance
        
    Returns
    -------
    complex
        S21 transmission coefficient (V_load / V_input)
    """
    n = len(L_array)
    w = 2.0 * np.pi * fc
    
    # Fixed attenuation factor (not dependent on load)
    att_factor = 10.0**(-input_atten_dB/20.0)  # Negative for attenuation
    
    # Step 1: Calculate impedance of each resonator
    Z_resonators = np.zeros(n, dtype=np.complex128)
    
    for i in range(n):
        # Parallel LC impedance (L_array already includes L_junk)
        if C_array[i] > 0:
            ZC = 1.0 / (1j * w * C_array[i])
            ZL = 1j * w * L_array[i]
            Z_parallel_inv = 1.0/ZC + 1.0/(ZL + R_array[i])
            Z_parallel = 1.0 / Z_parallel_inv
        else:
            Z_parallel = 1j * w * L_array[i] + R_array[i]
        
        # Add coupling capacitor (L_junk already included in L_array)
        ZCc = 1.0 / (1j * w * Cc_array[i])
        Z_resonators[i] = Z_parallel + ZCc
    
    # Step 2: Combine all resonators in parallel
    # 1/Z_total = sum(1/Z_i) for parallel impedances
    Z_total_inv = 0.0 + 0.0j
    for i in range(n):
        Z_total_inv += 1.0 / Z_resonators[i]
    
    # Avoid division by zero - if no resonators, use very high impedance
    if abs(Z_total_inv) > 1e-12:
        Z_total_resonators = 1.0 / Z_total_inv
    else:
        Z_total_resonators = 1e12 + 0j  # Very high impedance (no loading)
    
    # Step 3: Simple transmission line model
    # The resonators shunt current away from the load
    # Transmission coefficient is the voltage divider ratio
    
    # When resonators are off-resonance (high Z), most signal reaches load
    # When resonators are on-resonance (low Z), they shunt signal away
    
    # Transmission coefficient: how much reaches the load
    # S21 = ZLNA / (ZLNA + Z_series) where Z_series represents series loss
    # But our resonators are in parallel, so they act as shunt admittance
    
    # The proper formula for shunt elements on a transmission line:
    # V_load/V_source = 1 / (1 + Z_line/Z_shunt) for matched line
    # For our case with parallel resonators shunting to ground:
    # S21 = Z_shunt / (Z_shunt + Z_line)
    
    # Simplified model: resonators in parallel with load
    # Total load seen = Z_total_resonators || ZLNA
    Z_eff = 1.0 / (1.0/Z_total_resonators + 1.0/ZLNA)
    
    # Transmission coefficient (assuming Z0 source impedance)
    Z0 = system_termination
    S21_raw = Z_eff / (Z_eff + Z0)
    
    # Apply fixed attenuation and gain
    S21 = S21_raw * att_factor * GLNA
    
    return S21


# ============================================================================
# Helper Functions
# ============================================================================

@jit(nopython=True, cache=True)
def current_factors(currents_array, Istar):
    """
    Calculate current-dependent Lk modification factors.
    
    Parameters
    ----------
    currents_array : ndarray (complex)
        Resonator currents
    Istar : float
        Characteristic current
        
    Returns
    -------
    ndarray
        Lk modification factors (1 + I^2/I*^2)
    """
    return 1.0 + (np.abs(currents_array)**2 / Istar**2)



"""Base dialog for network analysis parameter entry."""

from .utils import (
    QtWidgets, QtCore, QRegularExpression, QRegularExpressionValidator,
    QDoubleValidator, QIntValidator,
    DEFAULT_AMPLITUDE, DEFAULT_MIN_FREQ, DEFAULT_MAX_FREQ, DEFAULT_CABLE_LENGTH,
    DEFAULT_NPOINTS, DEFAULT_NSAMPLES, DEFAULT_MAX_CHANNELS, DEFAULT_MAX_SPAN,
    DEFAULT_AMP_START, DEFAULT_AMP_STOP, DEFAULT_AMP_ITERATIONS,
    UnitConverter, traceback
)
from .tasks import DACScaleFetcher
import numpy as np  # For linspace

class NetworkAnalysisDialogBase(QtWidgets.QDialog):
    """
    Base class for network analysis dialogs, providing shared functionality
    for amplitude input (normalized and dBm), DAC scale handling, and
    parameter parsing.
    """
    def __init__(self, parent: QtWidgets.QWidget = None, params: dict = None,
                 modules: list[int] = None, dac_scales: dict[int, float] = None):
        """
        Initializes the base dialog.

        Args:
            parent: The parent widget.
            params: Dictionary of existing parameters to populate fields.
            modules: List of module numbers relevant to this dialog.
            dac_scales: Dictionary mapping module numbers to their DAC scales in dBm.
        """
        super().__init__(parent)
        self.params = params or {}  # Store initial parameters, default to empty dict
        self.modules = modules or [1, 2, 3, 4] # Default or passed-in modules
        # Initialize DAC scales for relevant modules, defaulting to None (unknown)
        self.dac_scales = dac_scales or {module_idx: None for module_idx in self.modules}
        self.currently_updating = False # Flag to prevent recursive updates between amp/dBm fields
        
    def setup_amplitude_group(self, layout: QtWidgets.QFormLayout) -> QtWidgets.QGroupBox:
        """
        Sets up the QGroupBox for amplitude settings (Normalized and dBm).

        This includes input fields for normalized amplitude and power in dBm,
        and a display for DAC scale information. Connections are made to
        synchronize and validate these fields.

        Args:
            layout: The QFormLayout to add the amplitude group to.

        Returns:
            The created QGroupBox containing amplitude settings.
        """
        amp_group = QtWidgets.QGroupBox("Amplitude Settings") # Group box title
        amp_layout = QtWidgets.QFormLayout(amp_group)
        
        # Determine initial amplitude: use 'amps' list if available, else 'amp', else default.
        amps_list = self.params.get('amps', [self.params.get('amp', DEFAULT_AMPLITUDE)])
        amp_str = ','.join(map(str, amps_list)) if amps_list else str(DEFAULT_AMPLITUDE)
        
        self.amp_edit = QtWidgets.QLineEdit(amp_str)
        self.amp_edit.setToolTip("Enter a single value or comma-separated list of normalized amplitudes (e.g., 0.001,0.01,0.1). Expressions like '1/1000' are allowed.")
        amp_layout.addRow("Normalized Amplitude:", self.amp_edit)
        
        self.dbm_edit = QtWidgets.QLineEdit()
        self.dbm_edit.setToolTip("Enter a single value or comma-separated list of power in dBm (e.g., -30,-20,-10). Expressions are allowed.")
        amp_layout.addRow("Power (dBm):", self.dbm_edit)
        
        self.dac_scale_info = QtWidgets.QLabel("Fetching DAC scales...")
        self.dac_scale_info.setWordWrap(True)
        amp_layout.addRow("DAC Scale (dBm):", self.dac_scale_info)
        
        # Connect signals for live updates (without validation) and validation on edit finish
        self.amp_edit.textChanged.connect(lambda: self._update_dbm_from_normalized(validate=False)) # Live update dBm field
        self.dbm_edit.textChanged.connect(lambda: self._update_normalized_from_dbm(validate=False)) # Live update normalized amp field
        self.amp_edit.editingFinished.connect(self._validate_normalized_values) # Validate on finishing edit
        self.dbm_edit.editingFinished.connect(self._validate_dbm_values)       # Validate on finishing edit

        # Linspace generator UI
        linspace_group = QtWidgets.QGroupBox("Generate Amplitude List")
        linspace_layout = QtWidgets.QFormLayout(linspace_group)

        self.start_amp_edit = QtWidgets.QLineEdit(f"{DEFAULT_AMP_START}")
        self.start_amp_edit.setValidator(QDoubleValidator(self))
        self.start_amp_edit.setToolTip("Start value for linspace generation.")
        linspace_layout.addRow("Start:", self.start_amp_edit)

        self.stop_amp_edit = QtWidgets.QLineEdit(f"{DEFAULT_AMP_STOP}")
        self.stop_amp_edit.setValidator(QDoubleValidator(self))
        self.stop_amp_edit.setToolTip("Stop value for linspace generation.")
        linspace_layout.addRow("Stop:", self.stop_amp_edit)

        self.iterations_amp_edit = QtWidgets.QLineEdit(f"{DEFAULT_AMP_ITERATIONS}")
        self.iterations_amp_edit.setValidator(QIntValidator(2, 1000, self)) # Min 2 points for linspace
        self.iterations_amp_edit.setToolTip("Number of points for linspace generation (min 2).")
        linspace_layout.addRow("Iterations:", self.iterations_amp_edit)
        
        button_layout = QtWidgets.QHBoxLayout()
        self.fill_amp_button = QtWidgets.QPushButton("Fill Normalized Amplitude")
        self.fill_amp_button.clicked.connect(self._on_fill_amplitude_clicked)
        button_layout.addWidget(self.fill_amp_button)

        self.fill_dbm_button = QtWidgets.QPushButton("Fill Power (dBm)")
        self.fill_dbm_button.clicked.connect(self._on_fill_dbm_clicked)
        button_layout.addWidget(self.fill_dbm_button)
        
        linspace_layout.addRow(button_layout)
        amp_layout.addRow(linspace_group) # Add this subgroup to the main amplitude layout
        
        layout.addRow("Amplitude Settings:", amp_group)
        return amp_group

    def _on_fill_amplitude_clicked(self):
        """Handles the 'Fill Amplitude' button click."""
        try:
            start = float(self.start_amp_edit.text())
            stop = float(self.stop_amp_edit.text())
            iterations = int(self.iterations_amp_edit.text())

            if iterations < 2:
                QtWidgets.QMessageBox.warning(self, "Input Error", "Iterations must be at least 2.")
                return

            values = np.linspace(start, stop, iterations)
            # Use a general format, good for typical normalized amplitudes
            self.amp_edit.setText(", ".join([f"{v:.6g}" for v in values])) 
        except ValueError:
            QtWidgets.QMessageBox.warning(self, "Input Error", "Invalid input for Start, Stop, or Iterations.")
        except Exception as e:
            QtWidgets.QMessageBox.critical(self, "Error", f"Could not generate amplitude list: {str(e)}")

    def _on_fill_dbm_clicked(self):
        """Handles the 'Fill dBm' button click."""
        try:
            start = float(self.start_amp_edit.text())
            stop = float(self.stop_amp_edit.text())
            iterations = int(self.iterations_amp_edit.text())

            if iterations < 2:
                QtWidgets.QMessageBox.warning(self, "Input Error", "Iterations must be at least 2.")
                return

            values = np.linspace(start, stop, iterations)
            # Use a format suitable for dBm values
            self.dbm_edit.setText(", ".join([f"{v:.2f}" for v in values]))
        except ValueError:
            QtWidgets.QMessageBox.warning(self, "Input Error", "Invalid input for Start, Stop, or Iterations.")
        except Exception as e:
            QtWidgets.QMessageBox.critical(self, "Error", f"Could not generate dBm list: {str(e)}")
        


    def _validate_normalized_values(self):
        """
        Validates the entered normalized amplitude values after editing is finished.
        Shows a warning dialog if values are outside typical ranges (e.g., > 1.0 or < 1e-4).
        """
        amp_text = self.amp_edit.text().strip()
        if not amp_text:
            return

        # DAC scale is not strictly needed for validating normalized amplitude against 0-1 range,
        # but good to have for consistency if other checks were dac_scale dependent.
        # dac_scale = self._get_selected_dac_scale()
        # if dac_scale is None: return # Or proceed with partial validation

        warnings_list = []
        normalized_values = self._parse_amplitude_values(amp_text)
        for norm_val in normalized_values:
            if norm_val > 1.0:
                warnings_list.append(f"Warning: Normalized amplitude {norm_val:.6f} > 1.0 (maximum)")
            elif norm_val < 1e-4: # Arbitrary small value warning threshold
                warnings_list.append(f"Warning: Normalized amplitude {norm_val:.6f} < 1e-4 (minimum recommended)")
        
        if warnings_list:
            self._show_warning_dialog("Normalized Amplitude Warning", warnings_list)

    def _validate_dbm_values(self):
        """
        Validates the entered dBm values after editing is finished.
        Shows a warning dialog if values exceed DAC scale or result in problematic
        normalized amplitudes.
        """
        dbm_text = self.dbm_edit.text().strip()
        if not dbm_text:
            return

        dac_scale = self._get_selected_dac_scale()
        if dac_scale is None:
            # Cannot validate dBm against DAC scale if unknown
            self._show_warning_dialog("DAC Scale Unknown", ["Cannot validate dBm values without a known DAC scale."])
            return

        warnings_list = []
        dbm_values = self._parse_dbm_values(dbm_text)
        for dbm_val in dbm_values:
            if dbm_val > dac_scale:
                warnings_list.append(f"Warning: {dbm_val:.2f} dBm > {dac_scale:+.2f} dBm (DAC maximum)")
            
            # Check corresponding normalized amplitude
            norm_val = UnitConverter.dbm_to_normalize(dbm_val, dac_scale)
            if norm_val > 1.0:
                warnings_list.append(f"Warning: {dbm_val:.2f} dBm results in normalized amplitude > 1.0 ({norm_val:.6f})")
            elif norm_val < 1e-4: # Arbitrary small value warning threshold
                warnings_list.append(f"Warning: {dbm_val:.2f} dBm results in normalized amplitude < 1e-4 ({norm_val:.6f})")
        
        if warnings_list:
            self._show_warning_dialog("dBm Amplitude Warning", warnings_list)

    def _show_warning_dialog(self, title: str, warnings_list: list[str]):
        """Displays a warning message box with a list of warnings."""
        QtWidgets.QMessageBox.warning(self, title, "\n".join(warnings_list))
            
    def _parse_numeric_values(self, text: str) -> list[float]:
        """
        Parses a comma-separated string of numeric values.
        Each part can be an expression evaluatable by `eval()`.
        Invalid parts are silently skipped.

        Args:
            text: The string containing numeric values (amplitude or dBm).

        Returns:
            A list of parsed float values.
        """
        values = []
        for part in text.split(','):
            part = part.strip()
            if part:
                try:
                    # Using eval allows for simple expressions like "1/1000".
                    # Caution: eval can execute arbitrary code if input is not controlled.
                    # In this GUI context, user inputs values for their own use.
                    values.append(float(eval(part)))
                except (ValueError, SyntaxError, NameError, TypeError):
                    # Silently skip parts that cannot be evaluated to a float
                    continue
        return values
    
    def _parse_amplitude_values(self, amp_text: str) -> list[float]:
        """Parse comma-separated amplitude values (delegates to _parse_numeric_values)."""
        return self._parse_numeric_values(amp_text)
        
    def _parse_dbm_values(self, dbm_text: str) -> list[float]:
        """Parse comma-separated dBm values (delegates to _parse_numeric_values)."""
        return self._parse_numeric_values(dbm_text)

    def _update_dac_scale_info(self):
        """
        Updates the DAC scale information label based on selected modules
        and their known DAC scales. Also enables/disables the dBm input field
        accordingly and triggers an update of dBm from normalized amplitude.
        """
        selected_modules = self._get_selected_modules()
        has_known_scale = False
        scales_text_list = []

        for module_idx in selected_modules:
            dac_scale = self.dac_scales.get(module_idx)
            if dac_scale is not None:
                has_known_scale = True
                scales_text_list.append(f"Module {module_idx}: {dac_scale:+.2f} dBm")
            else:
                scales_text_list.append(f"Module {module_idx}: Unknown")
        
        text_to_display = "\n".join(scales_text_list) if selected_modules else "Unknown (no modules selected)"
        self.dac_scale_info.setText(text_to_display)
        
        if has_known_scale:
            self.dbm_edit.setEnabled(True)
            self.dbm_edit.setToolTip("Enter dBm values (e.g., -30,-20,-10). Expressions are allowed.")
            self._update_dbm_from_normalized() # Update dBm field now that scale might be known/changed
        else:
            self.dbm_edit.setEnabled(False)
            self.dbm_edit.setToolTip("DAC scale unknown for selected module(s) - dBm input disabled.")
            self.dbm_edit.clear()
    
    def _get_selected_modules(self) -> list[int]:
        """
        Placeholder method to get the list of currently selected modules.
        Subclasses must override this to provide actual module selection logic.

        Returns:
            An empty list. Subclasses should return a list of integer module IDs.
        """
        # This method must be implemented by subclasses
        return [] 
        
    def _get_selected_dac_scale(self) -> float | None:
        """
        Retrieves the DAC scale for the currently selected module(s).
        If multiple modules are selected, it returns the DAC scale of the first
        module in the selection that has a known DAC scale.

        Returns:
            The DAC scale in dBm as a float, or None if no scale is known
            for any selected module or if no modules are selected.
        """
        selected_modules = self._get_selected_modules()
        if not selected_modules:
            return None
        
        for module_idx in selected_modules:
            dac_scale = self.dac_scales.get(module_idx)
            if dac_scale is not None:
                return dac_scale # Return the first known DAC scale
        return None # No known DAC scale for any of the selected modules
    
    def _update_dbm_from_normalized(self, validate: bool = True):
        """
        Updates the dBm field based on the normalized amplitude field.
        
        Args:
            validate: If True, validate values and show warnings (if field doesn't have focus).
                      If False, perform conversion without validation (for live updates).
        """
        if self.currently_updating or not self.dbm_edit.isEnabled():
            return
        self.currently_updating = True
        try:
            amp_text = self.amp_edit.text().strip()
            if not amp_text:
                self.dbm_edit.setText("")
                return

            dac_scale = self._get_selected_dac_scale()
            if dac_scale is None:
                self.dbm_edit.setEnabled(False)
                self.dbm_edit.setToolTip("Unable to query DAC scale for conversion.")
                self.dbm_edit.clear()
                return

            normalized_values = self._parse_amplitude_values(amp_text)
            dbm_values = []
            warnings_list = []
            
            for norm_val in normalized_values:
                # Collect validation warnings only if validate=True
                if validate:
                    if norm_val > 1.0:
                        warnings_list.append(f"Warning: Normalized amplitude {norm_val:.6f} > 1.0 (maximum)")
                    elif norm_val < 1e-4:
                        warnings_list.append(f"Warning: Normalized amplitude {norm_val:.6f} < 1e-4 (minimum recommended)")
                dbm_values.append(f"{UnitConverter.normalize_to_dbm(norm_val, dac_scale):.2f}")
            
            self.dbm_edit.setText(", ".join(dbm_values))
            
            # Show warnings only if validate=True and field doesn't have focus
            if validate and warnings_list and not self.amp_edit.hasFocus():
                self._show_warning_dialog("Normalized Amplitude Warning", warnings_list)
        finally:
            self.currently_updating = False
    
    def _update_normalized_from_dbm(self, validate: bool = True):
        """
        Updates the normalized amplitude field based on the dBm field.
        
        Args:
            validate: If True, validate values and show warnings (if field doesn't have focus).
                      If False, perform conversion without validation (for live updates).
        """
        if self.currently_updating or not self.dbm_edit.isEnabled():
            return
        self.currently_updating = True
        try:
            dbm_text = self.dbm_edit.text().strip()
            if not dbm_text:
                self.amp_edit.setText("")
                return

            dac_scale = self._get_selected_dac_scale()
            if dac_scale is None:
                # This state should ideally be prevented by disabling dbm_edit.
                # If somehow reached, cannot convert.
                self.amp_edit.setToolTip("Unable to query DAC scale for conversion.")
                # self.amp_edit.clear() # Avoid clearing if user is typing in dBm field
                return

            dbm_values = self._parse_dbm_values(dbm_text)
            normalized_values = []
            warnings_list = []
            
            for dbm_val in dbm_values:
                # Collect validation warnings only if validate=True
                if validate:
                    if dbm_val > dac_scale:
                        warnings_list.append(f"Warning: {dbm_val:.2f} dBm > {dac_scale:+.2f} dBm (DAC maximum)")
                
                norm_val = UnitConverter.dbm_to_normalize(dbm_val, dac_scale)
                
                # Validate resulting normalized amplitude only if validate=True
                if validate:
                    if norm_val > 1.0:
                        warnings_list.append(f"Warning: {dbm_val:.2f} dBm results in normalized amplitude > 1.0 ({norm_val:.6f})")
                    elif norm_val < 1e-4:
                        warnings_list.append(f"Warning: {dbm_val:.2f} dBm results in normalized amplitude < 1e-4 ({norm_val:.6f})")
                normalized_values.append(f"{norm_val:.6f}")
            
            self.amp_edit.setText(", ".join(normalized_values))

            # Show warnings only if validate=True and field doesn't have focus
            if validate and warnings_list and not self.dbm_edit.hasFocus():
                self._show_warning_dialog("dBm Amplitude Warning", warnings_list)
        finally:
            self.currently_updating = False



# Copyright (c) 2024 t0.technology, Inc.
#
# Distributed under the terms of the BSD license.  The full license is in the
# file LICENSE, distributed as part of this software.

import ast
import textwrap
import IPython


class StacktraceTidyTransformer(ast.NodeTransformer):
    """
    Overwrite all source-code metadata in an AST, including children.

    This is a bigger version of ast.copy_location, which modifies only a single
    node and not its children. We need it to preserve stack-trace data even
    when we insert new nodes in an AST.
    """

    def __init__(self, source):
        self._source = source

    def visit(self, node):
        return self.generic_visit(ast.copy_location(node, self._source))


# Convert coroutines in assignments into tasks
class CoroutineTransformer(ast.NodeTransformer):
    """
    Convert coroutine() objects in Expr and Assignment nodes into Tasks.
    """

    def visit_Module(self, node):
        """
        Insert "import" statements that we'll need to use later
        """

        node.body.insert(
            0,
            ast.Import(
                names=[
                    ast.alias(name="inspect", asname="_awaitless_inspect"),
                    ast.alias(name="asyncio", asname="_awaitless_asyncio"),
                ]
            ),
        )
        return self.generic_visit(node)

    def visit_FunctionDef(self, node):
        """
        Inhibit AST transformation when we bump into Function nodes

        This inhibition is mandatory, because our AST modifications insert
        "await" nodes that are syntactically illegal inside a synchronous
        function context.
        """

        # subtree walk is inhibited without generic_visit(node)
        return node

    def visit_AsyncFunctionDef(self, node):
        """
        Inhibit AST transformation inside AsyncFunctionDef nodes.

        This is a judgement call - we need to draw a boundary between "I'm in an
        interactive shell", where we do AST transformations, and "I'm in a
        codebase" where we don't.

        Because we are unable to do AST transformations on synchronous
        functions (see visit_FunctionDef above), and because writing functions
        seems like a "coding" and not an "interacting" activity, it seems
        reasonable to draw the boundary at function calls of both synchronous
        and asynchronous types.
        """

        return node

    def visit_ClassDef(self, node):
        """
        Inhibit AST transformation inside ClassDef nodes.

        We need to prevent assignments inside class bodies from being
        transformed, since an "await" there is illegal. (Functions defined in
        class bodies are already covered.)
        """

        return node

    def visit_Expr(self, node):
        """
        Rewrite Expr nodes
        """

        t = ast.parse(
            textwrap.dedent(
                """
                    _awaitless_ttemp = None
                    if _awaitless_inspect.iscoroutine(_awaitless_ttemp):
                        _awaitless_ttemp = _awaitless_asyncio.create_task(_awaitless_ttemp)
                        await _awaitless_ttemp
                    _awaitless_ttemp
                """
            )
        )

        t.body[0].value = node.value  # assign _awaitless_ttemp

        # Replace node and don't generic_visit its children
        t = StacktraceTidyTransformer(node).visit(t)
        return t.body

    def visit_Assign(self, node):
        """
        Rewrite Assign nodes
        """

        t = ast.parse(
            textwrap.dedent(
                """
                    _awaitless_ttemp = None
                    if _awaitless_inspect.iscoroutine(_awaitless_ttemp):
                        _awaitless_ttemp = _awaitless_asyncio.create_task(_awaitless_ttemp)
                        await _awaitless_ttemp
                    target = _awaitless_ttemp
                """
            )
        )

        t.body[0].value = node.value  # assign _awaitless_ttemp
        t.body[-1].targets = node.targets  # assign results

        # Replace node and don't generic_visit its children
        t = StacktraceTidyTransformer(node).visit(t)
        return t.body


def load_ipython_extension(ipython=None):
    if ipython is None:
        ipython = IPython.get_ipython()

    if ipython is None:
        raise RuntimeError("Can't install awaitless: no IPython session exists!")

    ast_transformers = ipython.ast_transformers
    ct = CoroutineTransformer()

    # If there's already an Awaitless instance installed, try to remove it -
    # otherwise we end up multiply transforming AST nodes, which explodes
    # very quickly
    for t in ast_transformers:
        if (
            t.__class__.__module__ == ct.__class__.__module__
            and t.__class__.__name__ == ct.__class__.__name__
        ):
            ast_transformers.remove(t)

    ast_transformers.append(ct)

    # We've created an implicit eventloop dependency - ensure all cells are
    # executed in an async context
    ipython.should_run_async = lambda *a, **kw: True



"""
This module is derived from and based on the mr_resonator project by Maclean Rouble:
https://github.com/macleaner/mr_resonator

Modifications in this repository include:
- Trimmed to include only functions used by the Mock CRS framework
- Consolidation and integration with local JIT physics approximations

Original project license: see rfmux/mr_resonator/LICENSE (upstream LICENSE retained)
"""

import numpy as np
from typing import Optional


class MR_LEKID():
    """
    Circuit-level LEKID model used by the mock framework.

    Model
    -----
    - Parallel RLC (R || jωL || 1/(jωC)) in series with coupling capacitor Cc,
      and optional parasitic inductance L_junk.
    - The network is driven through a p-type, 3-resistor attenuator synthesized for Z0=50 Ω.
    - The LNA is modeled as a complex input impedance ZLNA with linear gain GLNA.

    Usage
    -----
    Instances are created by MR_complex_resonator to carry circuit parameters
    (R, Lk, Lg, C, Cc, etc.). Methods such as compute_Vout() and compute_fr()
    are called by MockResonatorModel during S21 evaluation and convergence.
    """
    
    def __init__(
        self,
        C: float = 1e-12,
        R: float = 1e-6,
        Cc: float = 5e-15,
        Lk: float = 1e-9,
        Lg: Optional[float] = None,
        alpha_k: float = 0.5,
        L_junk: float = 0.0,
        Qi: Optional[float] = None,
        Qc: Optional[float] = None,
        Vin: Optional[float] = None,
        fr_presign2: int = -1,
        system_termination: float = 50.0,
        input_atten_dB: float = 20.0,
        ZLNA: complex = complex(50.0, 0.0),
        GLNA: float = 1.0,
        name: str = 'LR SERIES',
        LNA_noise_temperature: float = 6.0,
        plot_response: bool = False,
        verbose: bool = False
    ):
        
            
        self.Lk = Lk
        if Lg is None:
            # If Lg not provided, derive it from alpha_k parameter
            # alpha_k = Lk / (Lk + Lg) => Lg = Lk * (1 - alpha_k) / alpha_k
            self.Lg = (self.Lk - alpha_k*self.Lk) / alpha_k
        else:
            self.Lg = Lg
        
        self.L_junk = L_junk
        # L is the total inductance affecting resonance frequency and kinetic inductance fraction
        # L_total = Lk + Lg + L_junk (all inductances in the resonator circuit)
        self.L = self.Lk + self.Lg + self.L_junk
        # Effective kinetic inductance fraction (diluted by geometric and junk inductance)
        self.alpha_k = self.Lk / self.L if self.L > 0 else 0.0

        self.C = C
        self.R = R
        self.Cc = Cc
        self.name = name
        
            
         # readout params
        self.system_termination = system_termination
        self.input_atten_dB = input_atten_dB
        self.ZLNA = ZLNA
        self.GLNA = GLNA # LNA gain
        if Vin is None:
            self.Vin = 1e-5 # arbitrary choice
        else:
            self.Vin = Vin
            
        # noise params
        self.LNA_noise_temperature = LNA_noise_temperature
        self.LNA_noise_vrms_per_rtHz = np.sqrt(1.38e-23 * self.LNA_noise_temperature * 4 * self.system_termination) # over a 1 Hz bw

        self.nonres_flag = False # if the imaginary part of the impedance has no real roots
        
        
            
        if verbose:
            print('Created new resonator, %s, with params:'%(self.name))
            #print('Created new resonator, %s, with params:\nLk=%.2e H, Lg=%.2e H, C=%.2e F, Cc=%.2e F, R=%.2e ohm.'%(self.name, self.Lk, self.Lg, self.C, self.Cc, self.R))
            print(self.generate_res_param_string())
        
        
    def parallel_RLC(self, fc, C=None, L=None, R=None):
        """
        Impedance of the parallel RLC subnetwork at a probe frequency.

        Parameters
        ----------
        fc : float
            Probe frequency [Hz]
        C : float, optional
            Capacitance [F]
        L : float, optional
            Total inductance L = Lk + Lg [H]
        R : float, optional
            Series loss resistance [Ω]

        Returns
        -------
        complex
            Parallel RLC impedance [Ω]
        """
        # where fc is the carrier frequency
        # Compute the impedance of the parallel RLC only
    
        if C is None:
            C = self.C
        if L is None:
            L = self.L
        if R is None:
            R = self.R
            
        w = 2*np.pi*fc
        ZC = 1./(1j*w*C)
        ZL = 1j*w*L

        return 1./(1./ZC + 1./(ZL + R))
    
    def total_impedance(self, fc, C=None, L=None, R=None, Cc=None):
        """
        Total device impedance at fc including series elements.

        Composition
        -----------
        Z_total = Z_parallel_RLC + Z_Cc
        
        Note: L_junk is included in self.L (self.L = Lk + Lg + L_junk) for the
        parallel RLC calculation, so it affects the resonance frequency and
        dilutes the kinetic inductance fraction (alpha_k).

        Parameters
        ----------
        fc : float
            Probe frequency [Hz]
        C, L, R : float, optional
            RLC parameters [F, H, Ω] (L already includes L_junk)
        Cc : float, optional
            Coupling capacitor [F]

        Returns
        -------
        complex
            Total device impedance [Ω]
        """
        if C is None:
            C = self.C
        if L is None:
            L = self.L
        if R is None:
            R = self.R
        if Cc is None:
            Cc = self.Cc
        
        Zres = self.parallel_RLC(fc, L=L, C=C, R=R)
        ZCc = 1./(1j*2*np.pi*fc*Cc)

        return Zres + ZCc
    
    
    
    def compute_Vout(self, fc, Vin=None, L=None, C=None, R=None, Cc=None, ZLNA=None, GLNA=None, input_atten_dB=None):
        """
        Compute LNA input voltage from a single LEKID resonator at probe frequency.

        Parameters
        ----------
        fc : float
            Probe frequency [Hz]
        Vin : float, optional
            Source voltage before the input attenuator [V]
        L, C, R, Cc : float, optional
            Circuit parameters [H, F, Ω, F] (defaults to self values)
        ZLNA : complex, optional
            Complex LNA input impedance [Ω]
        GLNA : float, optional
            LNA gain (linear)
        input_atten_dB : float, optional
            Input attenuator value [dB]

        Returns
        -------
        complex
            Vout at the LNA input after GLNA [V]

        Notes
        -----
        - Attenuator is modeled as a p-type 3-resistor network synthesized for Z0=50 Ω.
        - The resonator is in parallel with ZLNA; the attenuator sits in series before
          that parallel combination.
        - This is the building block used by S21 calculations in the mock framework.
        """
        # Model the resonator as a voltage divider between it in parallel
        # with a complex input impedance of the LNA and an input 50 ohm attenuator,
        # return the voltage of the carrier across the resonator at a given frequency
        # as Vout, given an input carrier voltage at that frequency before the attenuator
        
        if C is None:
            C = self.C
        if L is None:
            L = self.L
        if R is None:
            R = self.R
        if Cc is None:
            Cc = self.Cc
        if Vin is None:
            Vin = self.Vin
        if ZLNA is None:
            ZLNA = self.ZLNA
        if GLNA is None:
            GLNA = self.GLNA
        if input_atten_dB is None:
            input_atten_dB = self.input_atten_dB
            
        r1, r2, r3 = self.get_att_vals(input_atten_dB)
        
        parallel = 1. / ( 1./self.total_impedance(fc, L=L, C=C, R=R, Cc=Cc) + 1./ZLNA )
        
        Vres = Vin * self.ptype(parallel, r1, r2, r3)
        Vout = GLNA * Vres
        return Vout
    

    def ptype(self, rl, r1=61.11, r2=247.5, r3=61.11):
        """
        P-type attenuator voltage division given a load.

        Parameters
        ----------
        rl : float or complex
            Load impedance [Ω]
        r1, r2, r3 : float, optional
            Attenuator resistor values synthesized for Z0 [Ω]

        Returns
        -------
        complex
            VL / Vin ratio at the load node (pre-LNA)
        """
        req = 1. / (1./r3 + 1./rl)
        VLoverVin = req / (req + r2)
        return VLoverVin


    
    def get_att_vals(self, att, z0=50.):
        """
        Compute p-type attenuator component values for target Z0 and attenuation.

        Parameters
        ----------
        att : float
            Attenuation [dB]
        z0 : float, optional
            System impedance [Ω]

        Returns
        -------
        tuple[float, float, float]
            (r1, r2, r3) in ohms
        """
        
        r1 = z0 * ((10**(att/20.) +1) / (10**(att/20.) - 1))
        r3 = r1
        r2 = (z0 / 2.) * ((10**(att/10.) - 1) / (10**(att/20.)))

        return r1, r2, r3
    
    def calc_Iin(self, fc, Vin=None, Zres=None):
        """
        Estimate input current into the attenuator-resonator-LNA network.

        Parameters
        ----------
        fc : float
            Probe frequency [Hz]
        Vin : float, optional
            Source voltage before attenuator [V]
        Zres : complex, optional
            Total device impedance at fc [Ω]

        Returns
        -------
        complex
            Input current [A]
        """
        if Vin is None:
            Vin = self.Vin
        if Zres is None:
            Zres = self.total_impedance(fc)
        r1, r2, r3 = self.get_att_vals(self.input_atten_dB)
        Zsys = 1. / ( 1./Zres + 1./self.ZLNA )
        Zp = 1. / ( 1./Zsys + 1./r3 )
        I2 = Vin / (r2 + Zp)
        Iin = I2 * ( r3 / (Zsys + r3) )
        return Iin
    
    def calc_Ires(self, fc, Zres=None, Iin=None, Vin=None, ZLNA=50., Z_other=None):
        """
        Estimate resonator branch current using a current divider approximation.

        Parameters
        ----------
        fc : float
            Probe frequency [Hz]
        Zres : complex, optional
            Resonator impedance at fc [Ω]
        Iin : complex, optional
            Input current [A]
        Vin : float, optional
            Source voltage before attenuator [V]
        ZLNA : complex or float, optional
            LNA input impedance [Ω]
        Z_other : complex, optional
            Optional additional shunt path [Ω]

        Returns
        -------
        complex
            Resonator branch current [A]

        Notes
        -----
        Approximation treats the last-stage attenuator resistor, resonator, and LNA
        input impedance as a three-way current divider. Sufficient for convergence loop.
        """
        
        if Zres is None:
            Zres = self.total_impedance(fc=fc)
        if Vin is None:
            Vin = self.Vin
        if Iin is None:
            Iin = self.calc_Iin(fc=fc, Zres=Zres, Vin=Vin)

        _, _, r3 = self.get_att_vals(self.input_atten_dB)
            
        if Z_other is not None:
            Zpar = 1./ ( 1./r3 + 1./Zres + 1./ZLNA + 1./Z_other )
        else:
            Zpar = 1./ ( 1./r3 + 1./Zres + 1./ZLNA)
        Ires = Iin * Zpar / Zres
        return Ires
    
    
    
    
    ############
    # L and fr #
    ############
    
    def compute_fr(self, L=None, C=None, R=None, Cc=None, presign2 = 1, verbose=False):
        """
        Estimate resonant frequency including series coupling capacitor.

        Method
        ------
        Uses a closed-form expression for Cc + (L+R || C). If no real solution is
        available (critically damped / non-resonant), falls back to searching for the
        magnitude minimum of Vout near the nominal 1/(2π√(LC)).

        Parameters
        ----------
        L, C, R, Cc : float, optional
            Circuit parameters [H, F, Ω, F]
        presign2 : int, optional
            Branch sign selector for the square root (kept for legacy compatibility)
        verbose : bool, optional
            Emit diagnostic prints

        Returns
        -------
        float
            Estimated resonant frequency [Hz]

        Notes
        -----
        Sets self.nonres_flag when the imaginary root test indicates no true resonance.
        """
        # x = -sqrt(-(C^2 R^2)/(2 (C^2 L^2 + C D L^2)) - (C D R^2)/(2 (C^2 L^2 + C D L^2)) - (i sqrt((i C^2 R^2 + i C D R^2 - 2 i C L - i D L)^2 - 4 i (i C^2 L^2 + i C D L^2)))/(2 (C^2 L^2 + C D L^2)) + (C L)/(C^2 L^2 + C D L^2) + (D L)/(2 (C^2 L^2 + C D L^2)))
        if C is None:
            C = self.C
        if L is None:
            L = self.L
        if R is None:
            R = self.R
        if Cc is None:
            Cc = self.Cc
        
        D = Cc
        presign = 1 # take only positive roots
        
        # check if we are going to end up with an imaginary solution:
        num3_part2 = 4 * (C**2 * L**2 + C * D * L**2)
        num3_part1 = (C**2 * R**2 + C * D * R**2 - 2 * C * L - D * L)**2
        if num3_part2 > num3_part1:
            # there is no real root to the imaginary part of the impedance
            # instead, look for the local minimum and call this the resonant frequency
            # though, arguable whether or not this is still a resonance
            
            self.nonres_flag = True
            if verbose:
                print('Found unreal solution! Looking for a local minimum instead.\nnum3 root arg: %.2e (num3 part 1: %.2e, num3 part 1: %.2e)'%(num3_part1-num3_part2, num3_part1, num3_part2))
            guess_fr = 1./(np.pi*2 * np.sqrt(L*C))
            guess_Q = 1./(2*np.pi*R*C)
            guess_bw = guess_fr/guess_Q
            
            span = guess_bw*10
            frange = np.linspace(guess_fr-span*2, guess_fr+span/2, 1000)
            test_mag = abs(self.compute_Vout(frange))
            guess_fr = frange[test_mag.argmin()]
            span2 = guess_bw
            frange = np.linspace(guess_fr-span, guess_fr+span, 1000)
            test_mag = abs(self.compute_Vout(frange))
            better_guess_fr = frange[test_mag.argmin()]
            return better_guess_fr
            
#             numerator3 = -1 * np.sqrt(abs((C**2 * R**2 + C * D * R**2 - 2 * C * L - D * L)**2 - 4 * (C**2 * L**2 + C * D * L**2)))
        else:
            numerator3 = -1 * np.sqrt((C**2 * R**2 + C * D * R**2 - 2 * C * L - D * L)**2 - 4 * (C**2 * L**2 + C * D * L**2))

        quotient1 = -1 * (C**2 * R**2) / (2 * (C**2 * L**2 + C * D * L**2))
        quotient2 = -1 *(C * D * R**2) / (2 * (C**2 * L**2 + C * D * L**2))
#         numerator3 = -1 * np.sqrt((C**2 * R**2 + C * D * R**2 - 2 * C * L - D * L)**2 - 4 * (C**2 * L**2 + C * D * L**2))
        denom3 = (2 * (C**2 * L**2 + C * D * L**2)) 
        quotient4 = (C * L)/(C**2 * L**2 + C * D * L**2)
        quotient5 = (D * L)/(2 * (C**2 * L**2 + C * D * L**2))

        x = presign * np.sqrt( quotient1 + quotient2 + presign2*(numerator3 / denom3) + quotient4 + quotient5 )

        return x / (2 * np.pi)
    



    def compute_Qc(self, Z0=50):
        """
        External coupling quality factor from coupling capacitor.

        Parameters
        ----------
        Z0 : float, optional
            Feedline characteristic impedance [Ω]

        Returns
        -------
        float
            Qc (dimensionless)
        """
        fr = self.compute_fr()
        Qc = (8 * self.C) / (self.Cc**2 * (2 * np.pi * fr * Z0) )
        return Qc

    def compute_Qi(self):
        """
        Internal quality factor from loss R, using L = Lk + Lg.

        Returns
        -------
        float
            Qi (here named Qr in the codebase; returned for compatibility)
        """
        fr = self.compute_fr()
        L = self.Lk + self.Lg
        Qr = np.pi * fr *2 * L / self.R
        return Qr

    # def compute_Qi(self):

    #     Qr = self.compute_Qr()
    #     Qc = self.compute_Qc()
    #     Qi = 1./(1./Qr - 1./Qc)
    #     return Qi
    def compute_Qr(self):
        """
        Loaded quality factor from internal and coupling Q.

        Returns
        -------
        float
            Qr (dimensionless)
        """
        Qi = self.compute_Qi()
        Qc = self.compute_Qc()
        Qr = 1./(1./Qi + 1./Qc)
        return Qr

    def compute_Q_values(self):
        Qr = self.compute_Qr()
        Qi = self.compute_Qi()
        Qc = self.compute_Qc()
        return Qr, Qi, Qc


    #####
    # extras
    #####
    
    def generate_res_param_string(self):
        res_param_string = 'Lk=%.2e H, Lg=%.2e H, C=%.2e F, Cc=%.2e F, R=%.2e ohm'%(self.Lk, self.Lg, self.C, self.Cc, self.R)
        return res_param_string
